source("R/getwhas100dataset.R")
source("R/SurvivalDataSet.R")
source('R/bootstraps_p.R')
source("R/coxfit.R")
source("R/wod_4_p.R")
source("R/wod_6.R")
source("R/wod_coxdevianceresiduals.R")
source("R/wod_coxld.R")
source("R/wod_coxmartingaleresiduals.R")


source("R/dbht_p.R")

source("R/wod_4hist_k_p.R")
source("R/wod_4biasedhist_k_p.R")
source("R/paod_t_onesided.R")

#' Extract the most outlying observations following a criteria based on concordance with parallel processing
#' @param surv.object An obect of type survival::Surv containing lifetimes and right-censoring status
#' @param covariate.data A data frame containing the data with covariate values for each individual
#' @param sod.method One of c("osd","bht","dbht","ld","martingale","deviance")
#' @param B The number of bootstrap samples generated only applicable for "bht" and "dbht" methods
#' @param B.N the number of observations in each bootstrap sample
#' @param max.outliers This parameter is only used for the "osd" method   
#' @param mc.cores The number of cores to execute the methods "bht" and "dbht"
#' 
#' @return A data.frame containing the most outlying observations sorted by outlying score
#' 
#' @export
bcsod <- function(surv.object, covariate.data, sod.method, B, B.N, max.outliers, mc.cores){
  
  ## load library dependencies
  library(stats)
  library(survival)
  if( Sys.info()['sysname']=='Windows' ){
    library(parallelsugar)
  }else{
    library(parallel)
  }
  
  
  
  if( ! is.Surv(surv.object) ){
    
    stop("Parameter \"surv.object\" must be an object of type \"survival::Surv\" ")
  }
  
  if( ! is.data.frame(covariate.data) ){
    
    stop("Parameter \"covariate.data\" must be of type \"data.frame\".")
  } 
  
  
  # obtain the number of records (individuals)
  N <- attr(surv.object,"dim")[1]
  
  t <- surv.object[1:N,1]
  s <- surv.object[1:N,2]
  # create a survival dataset from the Surv object and data
  surv_dataset <- SurvivalDataSet(data = covariate.data ,time_vector = t, status_vector = s)
  
  
  if( sod.method=="osd" ){
    outlier_set <- wod_6(dataset = surv_dataset, max_outliers = max.outliers )
    return( outlier_set)
  }
  
  if( sod.method=="bht" ){  
    
    outlier_set <- mclapply(X = 1:N,FUN = wod_4_p, s=surv.object, covariate.data=covariate.data, B=B , B.N=B.N , mc.cores = mc.cores)
    ## unlist to matrix and sort by pvalue
    outlier_set_flat <-  matrix(unlist(outlier_set), ncol = 4, byrow = TRUE) 
    colnames(outlier_set_flat) <- c("obs_id","avg_delta","max_delta","pvalue")
    set_sorted  <- outlier_set_flat[order(outlier_set_flat[,4]),] 
    
    return(set_sorted)
  }

  if(sod.method=="dbht" ){    
    
    outlier_set <- mclapply(X = 1:N, FUN=dbht_p, s=surv.object ,covariate.data = covariate.data , B=B, B.N=B.N , mc.cores=mc.cores )
    outlier_set_flat <-  matrix(unlist(outlier_set), ncol = 2, byrow = TRUE) 
    colnames(outlier_set_flat) <- c("obs_id","pvalue")
    set_sorted  <- outlier_set_flat[order(outlier_set_flat[,2]),]
    return(set_sorted)
  }
  
  if( sod.method=="ld" ){
    outlier_set <- wod_coxld(dataset = surv_dataset)
    return(outlier_set)
    
  }
  
  if( sod.method=="martingale"){
    outlier_set <- wod_coxmartingaleresiduals(dataset = surv_dataset)
    return(outlier_set)
  }
  
  if( sod.method=="deviance"){
    outlier_set <- wod_coxdevianceresiduals(dataset = surv_dataset)
    return(outlier_set)
  }
  
  stop("Parameter \"sod.method\" must be one of c(\"osd\",\"bht\",\"dbht\",\"ld\",\"martingale\",\"deviance\") ")
  
}
